﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Vintagestory.API.Client;
using Vintagestory.API.Server;
using Vintagestory.API.Common;
using Vintagestory.GameContent;
using Vintagestory.API.Config;
using Vintagestory.API.MathTools;
using Vintagestory.API.Common.Entities;
using ProtoBuf;

namespace vsmovemod.src
{
    [ProtoContract(ImplicitFields = ImplicitFields.AllPublic)]
    public class NetworkApiTestMessage
    {
        public bool messagefromserver;
        public bool sprintStep;
        public bool walkStep;
        public bool noStep;
    }

    [ProtoContract(ImplicitFields = ImplicitFields.AllPublic)]
    public class NetworkApiTestResponse
    {
        public bool response;
    }

    public class MoveMod : ModSystem
    {
        private ICoreServerAPI api;
        private ICoreClientAPI capi;

        IServerNetworkChannel serverChannel;
        IClientNetworkChannel clientChannel;

        // ==== Variables ==== 
        public int maxChargeJump = 100;
        public int chargeJump = 0;

        // Player
        public string clientUID;

        // Counter
        public int maxDashCountdown = 15; // Ticks (50ms) to reset dash & slide ability --- min 6 !
        public int dashCountdown = 0;
        public int soundGlidingCountdown = 0;
        public int soundGlidingOverWaterCd = 0;
        public int soundClimbinglCountdown = 0;
        public int generalAudioDelay = 0;
        public int fallDamageDelay = 0;

        public int dashTime = 5;
        public int slideTime = 10;
        public int gliderDelay = 0;

        // Stamina & Stamina-Drain
        public static float maxStamina = 100; // Maximum stamina
        public static float stamina = maxStamina;

        public float drainStaminaClimbing = 1f; // Stamina drain while climbing
        public float drainStaminaChargeJump = 0; // Stamina drain for charge jump / 10 ( => 0.04 = 40% ) -- (float)(maxStamina * 0.03);
        public float drainStaminaDash = (float)(maxStamina * 0.33333); // Stamina drain for dash / slide / walljump
        public float drainStaminaHanging = 0.4f; // Stamina drain while hanging
        public float drainStaminaGliding = 0.025f; // Stamina drain while gliding 0.025f default
        public float drainStaminaLift = (float)(maxStamina * 0.001); // Additional stamina drain when using glider-liftup
        public float drainStaminaIdleModificator = 0.2f; // Stamina drain modificator while hanging / climbing idle (not moving) in %

        public int maxStaminaRegenDelay = 60; // Ticks (50ms) until stamina-regeneration
        public int staminaRegenDelay = 0;
        public int staminaRegenValue = 2; // Stamina regen per tick

        // Bools
        public static bool playerIsClimbing = false;
        public bool isCrawling = false;
        public bool isFalling = false;

        public bool staminaRegenIsDelayed = false;
        public bool isClimbable = false;
        public bool isHanging = false;
        public bool motionStopedAfterDash = true;
        public bool letGoClimbing = false;



        // Flight
        bool activateFlightComputer = false;
        bool wearingGliderT1 = false;
        public int flightComputer_refresh = 10;

        public int lastDurability;

        public int refreshE_pot = 0;
        public int refresh_lastX = 0;
        public int refresh_lastZ = 0;
        public int landingAnimBuffer = 0;
        public int timeLanded = 0;

        public bool inGliderMode = false;
        public bool hasGlided = false;
        public bool glideAnimToggled = false;
        public bool nearGround;
        public bool liftGreaterGravity = false;
        public bool flightIsStalling = false;
        public bool inert = false;
        public bool hasPlayedControlSound = false;
        public bool playerGlidingOverWater = false;

        public double F_gravity = 1;
        public double E_pot = 1;
        public double F_lift;
        public double F_drag;

        public double surface_liftBase = 2;  // 2 base
        public double surface_lift;
        public double surface_dragBase = 2;  // 2 base
        public double surface_drag;

        public double temperature;
        public double modTemperature;

        public double windSpeed;
        public double modWindSpeed;

        public double motionSumXZ = 1;

        double lastX;
        double lastZ;

        public int gliderUnfoldTimer = 0; // unused
        public int gliderInertTimer = 0; // unused

        // Debug bools
        public bool getFallDamage = true;

        // Hotkey bools
        public static bool toggleStandup = false;
        public static bool toggleCrawl = false;
        public static bool toggleGlider = false;

        // CMD
        public static bool noStep = false;
        public static bool sprintStep = true;
        public static bool walkStep = false;

        // Misc
        public static SimpleParticleProperties myParticles = new SimpleParticleProperties(1, 1, ColorUtil.ColorFromRgba(255, 255, 255, 1), new Vec3d(), new Vec3d(), new Vec3f(), new Vec3f());
        public static SimpleParticleProperties mySpecialParticles = new SimpleParticleProperties(1, 1, ColorUtil.ColorFromRgba(255, 255, 255, 1), new Vec3d(), new Vec3d(), new Vec3f(), new Vec3f());
        public static SimpleParticleProperties myWaterSplashParticles = new SimpleParticleProperties(1, 1, ColorUtil.ColorFromRgba(255, 255, 255, 1), new Vec3d(), new Vec3d(), new Vec3f(), new Vec3f());

        // Experimental
        bool slotChanged = false;

        // Item Initialisation
        public override void Start(ICoreAPI api)
        {
            base.Start(api);
            api.RegisterItemClass("ItemGlider", typeof(ItemGlider));
        }

        // SERVERSIDE =======================================================================================================================================================================================================================================
        public override void StartServerSide(ICoreServerAPI api)
        {
            this.api = api;

            api.Event.PlayerJoin += entity =>
            entity.Entity.Properties.FallDamage = true;

            serverChannel =
                api.Network.RegisterChannel("networkToggleGlider")
                .RegisterMessageType(typeof(NetworkApiTestMessage))
                .RegisterMessageType(typeof(NetworkApiTestResponse))
                .SetMessageHandler<NetworkApiTestResponse>(OnClientMessage)
            ;

            api.RegisterCommand("autostep", "Send a test network message", "", OnNwTestCmd, Privilege.controlserver);
            api.RegisterCommand("sprint-step", "Send a test network message", "", SprintStep, Privilege.controlserver);
            api.RegisterCommand("walk-step", "Send a test network message", "", WalkStep, Privilege.controlserver);
            api.RegisterCommand("no-step", "Send a test network message", "", NoStep, Privilege.controlserver);
        }

        private void SprintStep(IServerPlayer player, int groupId, CmdArgs args)
        {
            serverChannel.BroadcastPacket(new NetworkApiTestMessage()
            {
                messagefromserver = false,
                sprintStep = true,
                walkStep = false,
                noStep = false,
        });
        }
        private void WalkStep(IServerPlayer player, int groupId, CmdArgs args)
        {
            serverChannel.BroadcastPacket(new NetworkApiTestMessage()
            {
                messagefromserver = false,
                sprintStep = false,
                walkStep = true,
                noStep = false,
        });
        }
        private void NoStep(IServerPlayer player, int groupId, CmdArgs args)
        {
            serverChannel.BroadcastPacket(new NetworkApiTestMessage()
            {
                messagefromserver = false,
                sprintStep = false,
                walkStep = false,
                noStep = true,
        });
        }

        private void OnNwTestCmd(IServerPlayer player, int groupId, CmdArgs args)
        {
            serverChannel.BroadcastPacket(new NetworkApiTestMessage()
            {
                messagefromserver = false,
            });
        }

        private void OnClientMessage(IPlayer fromPlayer, NetworkApiTestResponse networkMessage)
        {
            fromPlayer.Entity.Properties.FallDamage = networkMessage.response;

            //api.SendMessageToGroup(
            //    GlobalConstants.GeneralChatGroup,
            //    "Network toggled falldamage to: " + fromPlayer.Entity.Properties.FallDamage.ToString() + " Response: " + networkMessage.response.ToString(),
            //    EnumChatType.Notification
            //);
        }

        // Called on Servercommand - sets falldamage = false;
        private void OnServerMessage(NetworkApiTestMessage networkMessage)
        {
            if (networkMessage.sprintStep)
            {
                sprintStep = true;
                walkStep = false;
                noStep = false;

                capi.ShowChatMessage("Autostep set to sprint");
            }
            else if (networkMessage.walkStep)
            {
                sprintStep = false;
                walkStep = true;
                noStep = false;
                capi.ShowChatMessage("Autostep set to walk");
            }
            else if (networkMessage.noStep)
            {
                sprintStep = false;
                walkStep = false;
                noStep = true;
                capi.ShowChatMessage("Autostep disabled");
            }

            //capi.ShowChatMessage("Sending response to client.");
            //clientChannel.SendPacket(new NetworkApiTestResponse()
            //{
            //    response = false
            //});
        }

        // OnServerMessage overload
        // Called on Gliding through air - sets falldamage = false;
        private void OnServerMessage()
        {
            clientChannel.SendPacket(new NetworkApiTestResponse()
            {
                response = false,
            });
        }

        // OnServerMessage overload
        // Called on leaving GliderMode - sets falldamage = false;
        private void OnServerMessageLeave()
        {
            clientChannel.SendPacket(new NetworkApiTestResponse()
            {
                response = true,
            });
        }

        public void SpawnFlightParticles(Entity byEntity)
        {
            myParticles.MinPos = byEntity.Pos.XYZ.Add(0, (byEntity.Properties.EyeHeight -1.5), 0).Ahead(-1.5f, byEntity.Pos.Pitch, byEntity.Pos.Yaw);
            myParticles.AddPos = new Vec3d(0.5, 0.5, 0.5);
            myParticles.LifeLength = 1F;
            myParticles.GravityEffect = 0.1F;
            myParticles.SizeEvolve = new EvolvingNatFloat(EnumTransformFunction.LINEAR, 4);
            myParticles.ParticleModel = EnumParticleModel.Quad;
            myParticles.Color = ColorUtil.ColorFromRgba(255, 255, 255, 30);
            myParticles.OpacityEvolve = new EvolvingNatFloat(EnumTransformFunction.LINEAR, -150);
            byEntity.World.SpawnParticles(myParticles);
        }

        public void SpawnSpecialParticles(Entity byEntity)
        {
            mySpecialParticles.MinPos = byEntity.Pos.XYZ.Add(0, 0.5, -0.5).Ahead(-0.5f, byEntity.Pos.Pitch, byEntity.Pos.Yaw);
            mySpecialParticles.AddPos = new Vec3d(1, 0.6, 1.2);
            mySpecialParticles.MaxSize = 0.001F;
            mySpecialParticles.LifeLength = 0.3F;
            mySpecialParticles.GravityEffect = -0.01F;
            mySpecialParticles.ParticleModel = EnumParticleModel.Cube;
            mySpecialParticles.MinQuantity = 0.2f;
            Random rand = new Random();
            mySpecialParticles.Color = ColorUtil.ColorFromRgba(255, rand.Next(0, 255), rand.Next(0, 255), rand.Next(0, 255));
            mySpecialParticles.OpacityEvolve = new EvolvingNatFloat(EnumTransformFunction.LINEAR, -255);
            byEntity.World.SpawnParticles(mySpecialParticles);
        }

        public void SpawnWaterSplashParticles(Entity byEntity)
        {
            myWaterSplashParticles.MinPos = byEntity.Pos.XYZ.Add(0, -0.3, 0).Ahead(-0.7f, byEntity.Pos.Pitch, byEntity.Pos.Yaw);
            myWaterSplashParticles.AddPos = new Vec3d(0, 0.1, 1.2);
            myWaterSplashParticles.MaxSize = 0.002F;
            myWaterSplashParticles.LifeLength = 1F;
            myWaterSplashParticles.GravityEffect = 0.01F;
            myWaterSplashParticles.ShouldDieInLiquid = true;
            myWaterSplashParticles.ParticleModel = EnumParticleModel.Cube;
            myWaterSplashParticles.MinQuantity = 1.5f;
            myWaterSplashParticles.Color = ColorUtil.ColorFromRgba(255, 255, 256, 30);
            myWaterSplashParticles.OpacityEvolve = new EvolvingNatFloat(EnumTransformFunction.LINEAR, -255);
            byEntity.World.SpawnParticles(myWaterSplashParticles);
        }

        // CLIENTSIDE =======================================================================================================================================================================================================================================
        public override void StartClientSide(ICoreClientAPI api)
        {
            this.capi = api;

            clientChannel =
                api.Network.RegisterChannel("networkToggleGlider")
                .RegisterMessageType(typeof(NetworkApiTestMessage))
                .RegisterMessageType(typeof(NetworkApiTestResponse))
                .SetMessageHandler<NetworkApiTestMessage>(OnServerMessage)
            ;

            //Spawn Init
            api.Event.PlayerEntitySpawn += entity =>
            entity.Entity.GetBehavior<EntityBehaviorControlledPhysics>().stepHeight = 0.9F;

            api.Event.PlayerEntitySpawn += entity =>
            entity.Entity.Properties.CanClimbAnywhere = false;

            // Hotkey Events
            // HotKey Stop crawling
            api.Input.RegisterHotKey("standup", "Stand Up from crawling", GlKeys.Space, HotkeyType.CharacterControls);
            api.Input.SetHotKeyHandler("standup", StandUp);

            // HotKey Start Crawl
            api.Input.RegisterHotKey("startcrawl", "Crawl on the floor using Sneak + Hotkey", GlKeys.R, HotkeyType.CharacterControls);
            api.Input.SetHotKeyHandler("startcrawl", StartCrawl);

            // HotKey Start Crawl
            api.Input.RegisterHotKey("useglider", "Use Glider when equiped / Let-go when climbing", GlKeys.F, HotkeyType.CharacterControls);
            api.Input.SetHotKeyHandler("useglider", UseGlider);

            // Cancel all animations while gliding
            // Game Time Tick Listener 1ms ========================================================
            api.Event.RegisterGameTickListener(dt2 =>
            {
                var player = api.World.Player;
                clientUID = player.PlayerUID;

                // Stop Animation when inGliderMode
                if (inGliderMode && hasGlided)
                {
                    player.Entity.StopAnimation("walk");
                    player.Entity.StopAnimation("idle");
                    player.Entity.StopAnimation("awaitdash");
                    player.Entity.StopAnimation("land");
                    player.Entity.StopAnimation("glide");
                    if (gliderDelay == 200)
                    {
                        // stop animation sprint a few moments after take off
                        player.Entity.StopAnimation("sprint");
                    }
                }

                // Stop Motion Vec AfterDash
                if (dashTime == 5 && motionStopedAfterDash == false && player.Entity.OnGround)
                {
                    Vec3d motionZero = new Vec3d(0, 0, 0);
                    player.Entity.Pos.Motion = motionZero;
                    motionStopedAfterDash = true;
                }

                // Prevent player from climbing down (bouldering)
                Block aboveBlock = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y + 1, (int)player.Entity.Pos.Z);
                Block aboveBlock2 = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y + 2, (int)player.Entity.Pos.Z);

                if (player.Entity.Controls.IsClimbing && !isClimbable && !aboveBlock.Climbable && !aboveBlock2.Climbable)
                {
                    player.Entity.Controls.Sneak = false;
                    api.Input.InWorldAction += Input_InWorldAction3;
                }
                else if (!isCrawling)
                {
                    api.Input.InWorldAction += Input_InWorldAction1;
                }

                // Toggle fall-damage when gliding / climbing
                if (inGliderMode || player.Entity.Controls.IsClimbing || isHanging)
                {
                    if (getFallDamage)
                    {
                        OnServerMessage();
                        getFallDamage = false;
                    }
                }
                else
                {
                    if (!getFallDamage)
                    {
                        fallDamageDelay++;
                        if (fallDamageDelay > 50)
                        {
                            OnServerMessageLeave();
                            fallDamageDelay = 0;
                            getFallDamage = true;
                        }
                    }
                }
            }, 1);

            // Glider stuff
            // Game Time Tick Listener 10ms ========================================================
            api.Event.RegisterGameTickListener(dt1 =>
            {
                var player = api.World.Player;

                //Check for glider wearable in slot
                var gearInv = player.Entity.GearInventory;

                foreach (var slot in gearInv)
                {
                    ItemWearable item;
                    if (slot.Empty || (item = slot.Itemstack.Collectible as ItemWearable) == null) continue; // null Exception catch
 
                    if (slot.Itemstack.Collectible.Code.Path.StartsWith("glider-body-"))
                    {
                        wearingGliderT1 = true;

                        if (slot.Itemstack.Collectible.Code.Path.StartsWith("glider-body-crafted-unfolded-rainbow") || 
                            slot.Itemstack.Collectible.Code.Path.StartsWith("glider-body-antique-unfolded-rainbow"))
                        {
                            SpawnSpecialParticles(player.Entity);
                        }
                    }
                    else
                    {
                        wearingGliderT1 = false;
                    }
                }

                // Array colors
                string[] colorArray = new string[11];
                colorArray[0] = "blue";
                colorArray[1] = "red";
                colorArray[2] = "green";
                colorArray[3] = "yellow";
                colorArray[4] = "plain";
                colorArray[5] = "black";
                colorArray[6] = "brown";
                colorArray[7] = "pink";
                colorArray[8] = "purple";
                colorArray[9] = "orange";
                colorArray[10] = "rainbow";

                // Swap Glider folded / unfolded
                foreach (var slot in gearInv)
                {
                    ItemWearable item;
                    if (slot.Empty || (item = slot.Itemstack.Collectible as ItemWearable) == null) continue; // null Exception catch

                    foreach (string color in colorArray)
                    {
                        if (inGliderMode && slot.Itemstack.Collectible.Code.Path.StartsWith("glider-body-crafted-folded-" + color))
                        {
                            Item gliderUnfolded = api.World.GetItem(new AssetLocation("vsmovemod:glider-body-crafted-unfolded-" + color));
                            ItemStack stackGliderUnfolded = new ItemStack(gliderUnfolded, 1);

                            slot.OnItemSlotModified(slot.Itemstack);
                            {
                                Item gliderFolded = api.World.GetItem(new AssetLocation("vsmovemod:glider-body-crafted-folded-" + color));
                                ItemStack stackGliderFolded = new ItemStack(gliderFolded, 1);

                                slot.Itemstack = stackGliderFolded;
                            }

                            slot.Itemstack = stackGliderUnfolded;
                            slot.MarkDirty();

                            api.World.PlaySoundFor(new AssetLocation("sounds/wearable/brigandine1"), player, false, 100, 5);
                        }
                        else if (!inGliderMode && timeLanded == 0 && slot.Itemstack.Collectible.Code.Path.StartsWith("glider-body-crafted-unfolded-" + color))
                        {
                            Item gliderFolded = api.World.GetItem(new AssetLocation("vsmovemod:glider-body-crafted-folded-" + color));
                            ItemStack stackGliderFolded = new ItemStack(gliderFolded, 1);

                            slot.OnItemSlotModified(slot.Itemstack);
                            {
                                slot.Itemstack = stackGliderFolded;
                            }

                            slot.Itemstack = stackGliderFolded;
                            slot.MarkDirty();

                            api.World.PlaySoundFor(new AssetLocation("sounds/wearable/brigandine2"), player, false, 100, 5);
                        }
                    }
                }        

                // lastX (Value-Abs.|Y|)
                refresh_lastX++;
                if (refresh_lastX == 10) // Refresh E_pot every 100 x 10ms = 1s
                {
                    if (!inert)
                    {
                        lastX = player.Entity.Pos.Motion.X;
                    }
                    refresh_lastX = 0;
                }

                // lastX (Value-Abs.|Y|)
                refresh_lastZ++;
                if (refresh_lastZ == 10) // Refresh E_pot every 100 x 10ms = 1s
                {
                    if (!inert)
                    {
                        lastZ = player.Entity.Pos.Motion.X;
                    }
                    refresh_lastZ = 0;
                }

                // Get Windspeed at playerPos
                windSpeed = player.Entity.World.BlockAccessor.GetWindSpeedAt(player.Entity.Pos.XYZ).X;

                // Get Temperatur at playerPos
                ClimateCondition getCLimateatPlayer = player.Entity.World.BlockAccessor.GetClimateAt(player.Entity.Pos.AsBlockPos);
                temperature = getCLimateatPlayer.Temperature;

                // Movement (=MotionSumXZ)
                if (motionSumXZ > 0)
                {
                    motionSumXZ = (Math.Abs(player.Entity.Pos.Motion.X) + Math.Abs(player.Entity.Pos.Motion.Z));
                }
                else // prevent 0 to prevent unlimited fallspeed
                {
                    motionSumXZ = (Math.Abs(player.Entity.Pos.Motion.X) + Math.Abs(player.Entity.Pos.Motion.Z)) + 0.001;
                }

                // E_pot (Value-Abs.|Y|)
                refreshE_pot++;

                if (refreshE_pot == 50) // Refresh E_pot every 100 x 10ms = 1s
                {
                    refreshE_pot = 0;

                    if (player.Entity.Pos.Motion.Y < 0)
                    {
                        double calcValueAbsYplus1 = 1 + (Math.Abs(player.Entity.Pos.Motion.Y)); // 1 - unlimited
                        E_pot = calcValueAbsYplus1;
                    }
                    else if (player.Entity.Pos.Motion.Y >= 0)
                    {
                        E_pot = 1;
                    }
                }

                // F_Drag
                //F_drag = motionSumXZ * modWindSpeed * Form_modTemperature (1.004 for 40C to 1.09 for -20 C ~ 9 %) * surface_drag;
                F_drag = (motionSumXZ * 0.6) * modWindSpeed * (2.09 - Math.Pow(modTemperature, (1.0 / 2.0))) * surface_drag;

                if (F_drag >= motionSumXZ) 
                {
                    flightIsStalling = true;
                }
                else
                {
                    flightIsStalling = false;
                }

                // F_Lift
                //F_lift = (motionSumXZ * modWindSpeed * modTemperature * surface_lift) + (E_pot * 0.015) => (min 1 * 0.015 => min F_lift == 0.015 for motionXZ = 0);
                F_lift = (motionSumXZ * modWindSpeed * modTemperature * surface_lift) + (E_pot * 0.015);
                
                if (F_lift > F_gravity)
                {
                    liftGreaterGravity = true;
                }
                else
                {
                    liftGreaterGravity = false;
                }

                // Temperature-Modificator
                modTemperature = Math.Pow((temperature + 21.0), (1.0 / 25.0)); // 1.0 to 1.17

                // Windspeed-Modificator
                if (windSpeed >= 0.1)
                {
                    modWindSpeed = Math.Pow(windSpeed, (1.0 / 3.0)); // 0.46 to 1.14
                }
                else
                {   
                    modWindSpeed = Math.Pow(0.1, (1.0 / 3.0)); // min windspeed 0.1 => min modWindSpeed 0.46
                }

                // Activate Glider-mode
                if (player.Entity.CollidedVertically && !player.Entity.Controls.IsClimbing && !isHanging && toggleGlider && !player.Entity.Controls.Jump && stamina > 0 && player.InventoryManager.ActiveHotbarSlot.Empty && wearingGliderT1)
                {
                    inGliderMode = true;
                    player.Entity.StartAnimation("readyglider");
                }

                // Glide when in air & Glider-mode on
                Block air = api.World.GetBlock(new AssetLocation("air"));
                Block below2Blocks = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y - 2, (int)player.Entity.Pos.Z);
                bool below2IsAir = (below2Blocks == air);

                if (inGliderMode && !player.Entity.CollidedVertically && !playerIsClimbing && !isHanging && stamina > 0 && below2IsAir && player.InventoryManager.ActiveHotbarSlot.Empty && wearingGliderT1)
                {
                    if (gliderDelay < 200) // Delay_Timer
                    {
                        gliderDelay++;
                    }

                    // Enable glider after min. Delay-elapsed
                    if (gliderDelay > 10)
                    {
                        // Set Properties
                        hasGlided = true;
                        stamina -= drainStaminaGliding;
                        player.Entity.Properties.SetEyeHeight(0.6);
                        api.Input.InWorldAction += Input_InWorldAction4;
                        player.Entity.StopAnimation("readyglider");
                        SpawnFlightParticles(player.Entity);

                        // Control player motion XZ
                        Vec3d walkVector = player.Entity.Controls.WalkVector * 0.3;  // Input speed modifier = increase for "powered version"

                        if (!player.Entity.Controls.Forward && !player.Entity.Controls.Backward && !player.Entity.Controls.Left && !player.Entity.Controls.Right && !player.Entity.Controls.Sprint && !player.Entity.Controls.Jump && gliderInertTimer < 50)
                        {
                            gliderInertTimer++;
                            double mod = 0.997;

                            player.Entity.Pos.Motion.X = lastX * mod;
                            lastX = lastX * mod;
                            player.Entity.Pos.Motion.X = lastZ * mod;
                            lastZ = lastZ * mod;
                        }
                        else
                        {
                            player.Entity.Pos.Motion.Add(walkVector * Math.Pow(E_pot, 3));
                            player.Entity.Pos.Motion.Sub(walkVector * F_drag);

                            gliderInertTimer = 0;
                        }

                        // Set bool inert
                        if (!(gliderInertTimer == 0) && !(gliderInertTimer == 50))
                        {
                            inert = true;
                        }
                        else
                        {
                            inert = false;
                        }

                        // Fallspeed (dY)
                        // Lift < Gravity
                        double var = 0.02; // modifier lift to deltaY

                        if (gliderDelay > 20)
                        {
                            if (!liftGreaterGravity)
                            {
                                player.Entity.Pos.Motion.Y = (-1 / F_lift) * var;
                            }
                            // Lift > Gravity
                            else
                            {
                                player.Entity.Pos.Motion.Y = (F_lift) * (var * 1);
                            }
                        }

                        // Increase Lift_surface (Liftup)
                        if (player.Entity.Controls.Jump && gliderDelay == 200)

                        {
                            stamina -= drainStaminaLift;

                            surface_lift = surface_liftBase * 3; // 2 -> 3 seems good
                            surface_drag = surface_dragBase * 3; // 3 -> 3 seems good
                        }
                        else
                        {
                            surface_lift = surface_liftBase;
                            surface_drag = surface_dragBase;
                        }

                        // Gliding animations =============================================================================================
                        if (!glideAnimToggled)
                        {
                            player.Entity.StartAnimation("useglider");
                            glideAnimToggled = true;
                        }

                        bool useGliderLeftToggle = false;
                        bool useGliderRightToggle = false;
                        bool useGliderBackToggle = false;
                        bool useGliderClimbToggle = false;
                        bool useGLiderFallToggle = false;

                        if (player.Entity.Controls.Left && !nearGround && landingAnimBuffer == 0)
                        {
                            if (!useGliderLeftToggle)
                            {
                                player.Entity.StopAnimation("useglider");
                                player.Entity.StopAnimation("usegliderright");
                                player.Entity.StopAnimation("usegliderback");
                                player.Entity.StopAnimation("usegliderfall");
                                player.Entity.StopAnimation("usegliderclimb");

                                player.Entity.StartAnimation("usegliderleft");
                                                         
                                useGliderLeftToggle = true;
                            }
                        }
                        else if (player.Entity.Controls.Right && !nearGround && landingAnimBuffer == 0)
                        {
                            if (!useGliderRightToggle)
                            {
                                player.Entity.StopAnimation("useglider");
                                player.Entity.StopAnimation("usegliderleft");
                                player.Entity.StopAnimation("usegliderback");
                                player.Entity.StopAnimation("usegliderfall");
                                player.Entity.StopAnimation("usegliderclimb");

                                player.Entity.StartAnimation("usegliderright");

                                useGliderRightToggle = true;
                            }
                        }
                        else if (player.Entity.Controls.Left && player.Entity.Controls.Right && !nearGround && landingAnimBuffer == 0)
                        {
                            if (!useGliderRightToggle && !useGliderRightToggle)
                            {
                                player.Entity.StopAnimation("useglider");
                                player.Entity.StopAnimation("usegliderright");
                                player.Entity.StopAnimation("usegliderleft");
                                player.Entity.StopAnimation("usegliderback");
                                player.Entity.StopAnimation("usegliderclimb");

                                player.Entity.StartAnimation("usegliderfall");

                                useGLiderFallToggle = true;
                            }
                        }
                        else if (player.Entity.Controls.Backward && !nearGround && landingAnimBuffer == 0)
                        {
                            if (!useGliderBackToggle)
                            {
                                player.Entity.StopAnimation("useglider");
                                player.Entity.StopAnimation("usegliderright");
                                player.Entity.StopAnimation("usegliderleft");
                                player.Entity.StopAnimation("usegliderfall");
                                player.Entity.StopAnimation("usegliderclimb");

                                player.Entity.StartAnimation("usegliderback");

                                useGliderBackToggle = true;
                            }
                        }
                        else if (player.Entity.Controls.Jump && !nearGround && landingAnimBuffer == 0)
                        {
                            if (!useGliderClimbToggle)
                            {
                                player.Entity.StopAnimation("useglider");
                                player.Entity.StopAnimation("usegliderright");
                                player.Entity.StopAnimation("usegliderleft");
                                player.Entity.StopAnimation("usegliderback");
                                player.Entity.StopAnimation("usegliderfall");

                                player.Entity.StartAnimation("usegliderclimb");

                                useGliderClimbToggle = true;
                            }
                        }
                        else
                        {
                            Block belowBlock7 = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y - 7, (int)player.Entity.Pos.Z);
                            Block belowBlock6 = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y - 6, (int)player.Entity.Pos.Z);
                            Block belowBlock5 = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y - 5, (int)player.Entity.Pos.Z);
                            Block belowBlock4 = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y - 4, (int)player.Entity.Pos.Z);
                            Block belowBlock3 = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y - 3, (int)player.Entity.Pos.Z);
                            Block belowBlock2 = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y - 2, (int)player.Entity.Pos.Z);
                            Block belowBlock1 = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y - 1, (int)player.Entity.Pos.Z);

                            // Spawn particles gliding close over water
                            if (belowBlock1.IsLiquid() || belowBlock2.IsLiquid() || belowBlock3.IsLiquid() || belowBlock4.IsLiquid() || belowBlock5.IsLiquid() || belowBlock6.IsLiquid())
                            {
                                SpawnWaterSplashParticles(player.Entity);
                                playerGlidingOverWater = true;
                            }
                            else
                            {
                                playerGlidingOverWater = false;
                            }

                            // Flight Anim
                            if (belowBlock1 == air && belowBlock2 == air && belowBlock3 == air && belowBlock4 == air && belowBlock5 == air && belowBlock6 == air && belowBlock7 == air)
                            {
                                if (!inert && landingAnimBuffer == 0)
                                {
                                    player.Entity.StopAnimation("usegliderleft");
                                    player.Entity.StopAnimation("usegliderright");
                                    player.Entity.StopAnimation("usegliderback");
                                    player.Entity.StopAnimation("usegliderclimb");
                                    player.Entity.StopAnimation("readytoland");

                                    if (player.Entity.Pos.Motion.Y > -0.1)
                                    {
                                        player.Entity.StopAnimation("usegliderfall");
                                        player.Entity.StartAnimation("useglider");

                                        useGliderLeftToggle = false;
                                        useGliderRightToggle = false;
                                        useGliderBackToggle = false;
                                        useGliderClimbToggle = false;
                                        useGLiderFallToggle = false;

                                        nearGround = false;
                                    }
                                    else
                                    {
                                        player.Entity.StopAnimation("useglider");
                                        player.Entity.StartAnimation("usegliderfall");

                                        useGliderLeftToggle = false;
                                        useGliderRightToggle = false;
                                        useGliderBackToggle = false;
                                        useGliderClimbToggle = false;
                                        useGLiderFallToggle = false;

                                        nearGround = false;
                                    }
                                }
                            }
                            else
                            {
                                if (gliderDelay == 200) // only play ready to land animation after glider Delay is maxed
                                {
                                    player.Entity.StopAnimation("useglider");
                                    player.Entity.StopAnimation("usegliderright");
                                    player.Entity.StopAnimation("usegliderleft");
                                    player.Entity.StopAnimation("usegliderback");
                                    player.Entity.StopAnimation("usegliderfall");
                                    player.Entity.StopAnimation("usegliderclimb");

                                    player.Entity.StartAnimation("readytoland");

                                    useGliderLeftToggle = false;
                                    useGliderRightToggle = false;
                                    useGliderBackToggle = false;
                                    useGliderClimbToggle = false;
                                    useGLiderFallToggle = false;

                                    nearGround = true;

                                    landingAnimBuffer = 50; // hold animation for at leat 50 * 10ms = 0.5s
                                }
                            }
                        }

                        // Animation Buffer Counter
                        if (landingAnimBuffer > 0)
                        {
                            landingAnimBuffer--;
                        }

                        // Gliding sounds passive
                        if (soundGlidingCountdown == 0)
                        {
                            Random rnd = new Random();
                            api.World.PlaySoundFor(new AssetLocation("sounds/wearable/leather1"), player, false, 100, 10);
                            soundGlidingCountdown = 25; // Delay Flight Sound
                        }
                        if (soundGlidingCountdown < 10 && !nearGround)
                        {
                            Random rnd = new Random();
                            api.World.PlaySoundFor(new AssetLocation("sounds/environment/wind"), player, false, 400, 50);
                            soundGlidingCountdown = 180;
                        }
                        if (soundGlidingCountdown < 180 && nearGround)
                        {
                            Random rnd = new Random();
                            api.World.PlaySoundFor(new AssetLocation("sounds/effect/swoosh"), player, false, 100, 10);
                            soundGlidingCountdown = rnd.Next(180, 240);
                        }
                        if (soundGlidingOverWaterCd < 20 && playerGlidingOverWater)
                        {
                            api.World.PlaySoundFor(new AssetLocation("vsmovemod:sounds/waterglide"), player, false, 100, 1000);
                            soundGlidingOverWaterCd = 40;
                        }

                        // Gliding sounds on Player-controls
                        if ((player.Entity.Controls.Backward || player.Entity.Controls.Left || player.Entity.Controls.Right || player.Entity.Controls.Jump) && !hasPlayedControlSound)
                        {
                            api.World.PlaySoundFor(new AssetLocation("sounds/effect/swoosh"), player, false, 100, 10);
                            hasPlayedControlSound = true;
                        }
                        if (!player.Entity.Controls.Backward && !player.Entity.Controls.Left && !player.Entity.Controls.Right && !player.Entity.Controls.Jump)
                        {
                            hasPlayedControlSound = false;
                        }
                    }  
                }

                // Cancel Glider-mode
                if ((inGliderMode && (stamina < 1 || !toggleGlider || !wearingGliderT1 || player.Entity.FeetInLiquid || player.Entity.Controls.IsClimbing || (hasGlided && player.Entity.CollidedVertically))) || (!hasGlided && player.Entity.Controls.Jump) || (inGliderMode && player.Entity.Controls.FloorSitting) || (inGliderMode && player.Entity.Controls.Sneak) || (inGliderMode && isCrawling) || (inGliderMode && !player.InventoryManager.ActiveHotbarSlot.Empty))
                {

                    api.Input.InWorldAction += Input_InWorldAction1;

                    // Land on ground sound
                    if (hasGlided && player.Entity.CollidedVertically)
                    {
                        api.World.PlaySoundFor(new AssetLocation("sounds/block/gravel"), player, false, 100, 10);
                    }

                    if (hasGlided)
                    {
                        timeLanded = 30; // Delay for folding wings
                    }

                    toggleGlider = false;
                    inGliderMode = false;
                    hasGlided = false;
                    gliderDelay = 0;
                    glideAnimToggled = false;

                    player.Entity.StopAnimation("useglider");
                    player.Entity.StopAnimation("usegliderright");
                    player.Entity.StopAnimation("usegliderleft");
                    player.Entity.StopAnimation("usegliderback");
                    player.Entity.StopAnimation("usegliderfall");
                    player.Entity.StopAnimation("usegliderclimb");
                    player.Entity.StopAnimation("readyglider");
                    player.Entity.StopAnimation("readytoland");

                    player.Entity.Properties.SetEyeHeight(1.7);
                    api.Input.InWorldAction += Input_InWorldAction1;
                }

                if (!hasGlided && timeLanded > 0)
                {
                    timeLanded--;
                }

            }, 10); //10 ms/tick

            // Game Time Tick Listener 50ms ========================================================
            api.Event.RegisterGameTickListener(dt =>
            {
                var player = api.World.Player;

                // DEBUG ==================
                //player.ShowChatNotification("Stamina " + stamina.ToString());
                //player.ShowChatNotification("Jumpforce " + chargeJump.ToString());
                //player.ShowChatNotification("Dash Timer: " + dashCountdown.ToString());
                //player.ShowChatNotification("Activity State: " + player.Entity.CurrentControls.ToString());
                //player.ShowChatNotification("Player Controls: " + player.Entity.Controls.ToString());
                //player.ShowChatNotification("Motion.Y: " + player.Entity.Pos.Motion.Y.ToString());
                //player.ShowChatNotification("Activity isClimbing?: " + player.Entity.Controls.IsClimbing.ToString());
                //player.ShowChatNotification("Activity Jump?: " + player.Entity.Controls.Jump.ToString());
                //player.ShowChatNotification("Block at PlayerPos cimbable?: " + isClimbable.ToString());
                //player.ShowChatNotification("Gravity per second: " + GlobalConstants.GravityPerSecond.ToString());
                //player.ShowChatNotification("Dash-Time: " + slideTime.ToString());
                //player.ShowChatNotification("Motion Stopped?: " + motionStopedAfterDash.ToString());
                //player.ShowChatNotification("Toggle Glider: " + toggleGlider.ToString());
                //player.ShowChatNotification("In Glider-Mode: " + inGliderMode.ToString());
                //player.ShowChatNotification("Glider-Delay: " + gliderDelay.ToString());
                //player.ShowChatNotification("Falldamage-Delay: " + fallDamageDelay.ToString());
                //player.ShowChatNotification("Get fall damage ?: " + player.Entity.Properties.FallDamage.ToString());
                //player.ShowChatNotification("Player on ground ? " + player.Entity.OnGround.ToString());
                //player.ShowChatNotification("Player collided? " + player.Entity.CollidedVertically.ToString());
                //player.ShowChatNotification("Last X: " + lastX.ToString());
                //player.ShowChatNotification("Last Z: " + lastZ.ToString());
                //player.ShowChatNotification("Inert: " + inert.ToString());
                //player.ShowChatNotification("Inert: " + gliderInertTimer.ToString());
                // DEBUG ==================

                // FLIGHT NAV-TOOLS           1
                flightComputer_refresh--;

                if (activateFlightComputer)
                {
                    if (flightComputer_refresh < 1)
                    {
                        player.ShowChatNotification("=========================================");

                        // CLIMATE CONDITIONS
                        //player.ShowChatNotification("Temperature: " + temperature.ToString());
                        //if (windSpeed != null)
                        //{
                        //    player.ShowChatNotification("Windspeed: " + windSpeed.ToString());
                        //}

                        player.ShowChatNotification("=========================================");

                        // FLIGHT FORCES
                        //player.ShowChatNotification("E_pot: " + E_pot.ToString());
                        //player.ShowChatNotification("F_gravity: " + F_gravity.ToString());
                        player.ShowChatNotification("F_lift: " + F_lift.ToString());
                        //player.ShowChatNotification("F_Drag: " + F_drag.ToString());

                        //FLIGHT SPEED
                        player.ShowChatNotification("Speed :" + motionSumXZ.ToString()); // SumAbsXZ
                        player.ShowChatNotification("Fall-Speed: " + player.Entity.Pos.Motion.Y.ToString());

                        //player.ShowChatNotification("Lift Surface: " + surface_lift.ToString());
                        //player.ShowChatNotification("Drag Surface: " + surface_drag.ToString());

                        flightComputer_refresh = 10;
                    }
                }

                // Get block at playerPos
                Block checkBlockatPlayerPos = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y, (int)player.Entity.Pos.Z);

                // Block above is air ?
                Block air = api.World.GetBlock(new AssetLocation("air"));
                Block aboveBlock = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y + 1, (int)player.Entity.Pos.Z);
                //player.ShowChatNotification("Block above: " + aboveBlock.ToString());

                // Counter per Tick (Dash Timer)
                if (dashCountdown > 0)
                {
                    dashCountdown--;
                }

                // Counter per Tick (Stamina Regen-Delay)
                if (staminaRegenDelay > 0)
                {
                    staminaRegenDelay--;
                }


                // Counter per Tick (FallDamage - Delay)
                if (fallDamageDelay > 0)
                {
                    fallDamageDelay--;
                }

                // Counter per Tick (SoundGliding - Countdown)
                if (soundGlidingCountdown > 0)
                {
                    soundGlidingCountdown--;
                }

                // Counter per Tick (SoundGliding - Countdown)
                if (soundGlidingOverWaterCd > 0)
                {
                    soundGlidingOverWaterCd--;
                }

                // Counter per Tick (SoundClimbing - Countdown)
                if (soundClimbinglCountdown > 0)
                {
                    soundClimbinglCountdown--;
                }

                // Counter per Tick (generalAudioDelay - Countdown)
                if (generalAudioDelay > 0)
                {
                    generalAudioDelay--;
                }

                //  0 Stamina Slow
                if (stamina < 1 && player.Entity.OnGround)
                {
                    Vintagestory.API.Config.GlobalConstants.OverallSpeedMultiplier = 0.75f;
                }
                else
                {
                    Vintagestory.API.Config.GlobalConstants.OverallSpeedMultiplier = 1f;
                }

                // Hanging
                Block aboveBlock2 = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y + 2, (int)player.Entity.Pos.Z);

                if (aboveBlock2 != air && stamina > 0 && !player.Entity.OnGround && player.Entity.Controls.Sneak && !player.Entity.Controls.TriesToMove && player.InventoryManager.ActiveHotbarSlot.Empty && isClimbable == false)
                {
                    Vec3d motionZero = new Vec3d(0, 0, 0);
                    player.Entity.Pos.Motion = motionZero;

                    player.Entity.Properties.Habitat = EnumHabitat.Air;
                    player.Entity.Properties.CanClimbAnywhere = false;
                    player.Entity.StopAnimation("hangingtop");
                    player.Entity.StartAnimation("hangingidle");
                    stamina -= (drainStaminaHanging * drainStaminaIdleModificator);
                    isHanging = true;
                }
                else if (aboveBlock2 != air && stamina > 0 && !player.Entity.OnGround && player.Entity.Controls.Sneak && player.Entity.Controls.TriesToMove && player.InventoryManager.ActiveHotbarSlot.Empty && isClimbable == false)
                {
                    player.Entity.Properties.Habitat = EnumHabitat.Air;
                    player.Entity.Properties.CanClimbAnywhere = false;
                    player.Entity.StopAnimation("hangingidle");
                    player.Entity.StartAnimation("hangingtop");
                    stamina -= drainStaminaHanging;
                    isHanging = true;
                }
                else if (aboveBlock2 != air && stamina > 0 && !player.Entity.OnGround && player.Entity.Controls.IsClimbing && player.InventoryManager.ActiveHotbarSlot.Empty && isClimbable == false)
                {
                    player.Entity.Properties.Habitat = EnumHabitat.Air;
                    player.Entity.Properties.CanClimbAnywhere = false;
                    player.Entity.StopAnimation("hangingidle");
                    player.Entity.StartAnimation("hangingtop");
                    stamina -= drainStaminaHanging;
                    isHanging = true;
                }
                else
                {
                    if (!inGliderMode)
                    {
                        player.Entity.Properties.Habitat = EnumHabitat.Land;
                        player.Entity.StopAnimation("hangingtop");
                        player.Entity.StopAnimation("hangingidle");
                        isHanging = false;
                    }
                }

                // Crawl funtion hitbox and animation
                if (isCrawling)
                {
                    player.Entity.Properties.CanClimbAnywhere = false;
                    player.Entity.StartAnimation("liegen");
                    player.Entity.SetHitbox(0.5f, 0.5f);
                    player.Entity.Properties.SetEyeHeight(0.6);
                    player.Entity.Properties.HitBoxSize = new Vec2f(0.5996094f, 0.9248045f);

                    if (player.Entity.Controls.TriesToMove)
                    {
                        player.Entity.StopAnimation("liegen");
                        player.Entity.StartAnimation("gleiten");
                    }
                    else
                    {
                        player.Entity.StopAnimation("gleiten");
                        player.Entity.StartAnimation("liegen");
                    }

                    // InputEventListener - Disable actions
                    api.Input.InWorldAction += Input_InWorldAction;

                    if (player.Entity.CurrentControls == ((EnumEntityActivity)8 | (EnumEntityActivity)1))
                    {
                        api.Input.InWorldAction += Input_InWorldAction2;
                    }
                    else if (player.Entity.CurrentControls == ((EnumEntityActivity)8 | (EnumEntityActivity)2))
                    {
                        api.Input.InWorldAction += Input_InWorldAction2;
                    }
                    else if (player.Entity.CurrentControls == (EnumEntityActivity)2)
                    {
                        player.Entity.Controls.StopAllMovement();
                        api.Input.InWorldAction += Input_InWorldAction;
                    }
                    else if (player.Entity.CurrentControls == ((EnumEntityActivity)4 | (EnumEntityActivity)2))
                    {
                        player.Entity.Controls.StopAllMovement();
                        api.Input.InWorldAction += Input_InWorldAction;
                    }
                    else
                    {
                        api.Input.InWorldAction += Input_InWorldAction;
                    }
                }

                // Stop crawling
                if (aboveBlock == air || aboveBlock.IsLiquid())
                {
                    if (((api.Input.MouseButton.Middle == true || toggleStandup == true) && dashCountdown < 1 && player.Entity.CollidedVertically && isCrawling == true) || (player.Entity.CurrentControls == (EnumEntityActivity.FloorSitting | EnumEntityActivity.Idle)))
                    {
                        isCrawling = false;
                        toggleStandup = false;
                        player.Entity.StopAnimation("gleiten");
                        player.Entity.StopAnimation("liegen");
                        player.Entity.Properties.SetEyeHeight(1.7);
                        player.Entity.Properties.HitBoxSize = new Vec2f(0.5996094f, 1.849609f);

                        // InputEventListener
                        api.Input.InWorldAction += Input_InWorldAction1;
                    }
                }

                //Reset toggleCrawl if not sneaking
                if (!player.Entity.Controls.Sneak)
                {
                    toggleCrawl = false;
                }

                //Start Crawl
                if ((player.Entity.Controls.Sneak && api.Input.MouseButton.Middle == true || player.Entity.Controls.Sneak && toggleCrawl == true) && player.Entity.CollidedVertically && !isCrawling && !(player.Entity.CurrentControls == (EnumEntityActivity.FloorSitting | EnumEntityActivity.Idle)))
                {
                    isCrawling = true;
                    toggleCrawl = false;
                }

                // Dash anticipation / better running Animation
                if ((player.Entity.CurrentControls == ((EnumEntityActivity)4 | (EnumEntityActivity)2)) && player.Entity.Controls.TriesToMove && dashCountdown < 1 && stamina > drainStaminaDash)
                {
                    player.Entity.StartAnimation("awaitdash");
                }
                else
                {
                    player.Entity.StopAnimation("awaitdash");
                }

                // Reset dash animation
                if (dashCountdown <= maxDashCountdown - 6)
                {
                    player.Entity.StopAnimation("dash");
                    player.Entity.StopAnimation("dashright");
                    player.Entity.StopAnimation("dashleft");
                }

                // Check if player is climbing for GUI
                if (player.Entity.Controls.IsClimbing)
                {
                    playerIsClimbing = true;
                }
                else
                {
                    playerIsClimbing = false;
                }

                // Dash start & animation
                if ((player.Entity.Controls.Sprint || player.Entity.Controls.Jump) && player.Entity.Controls.TriesToMove && player.Entity.Controls.RightMouseDown && !hasGlided)
                {
                    if (stamina > drainStaminaDash && dashCountdown < 1 && !player.Entity.Controls.Backward && !player.Entity.Controls.Right && !player.Entity.Controls.Left)
                    {
                        player.Entity.StartAnimation("dash");
                        stamina = stamina -= drainStaminaDash;
                        dashCountdown = maxDashCountdown;
                        staminaRegenDelay = maxStaminaRegenDelay;
                        dashTime = 0;
                        motionStopedAfterDash = false;

                        Random rnd = new Random();
                        int rndInt;
                        rndInt = rnd.Next(1, 3);
                        if (rndInt == 1)
                        {
                            api.World.PlaySoundFor(new AssetLocation("vsmovemod:sounds/dashforward-0"), player, false, 1.1f, 0.2f);
                        }
                        else if (rndInt == 2)
                        {
                            api.World.PlaySoundFor(new AssetLocation("vsmovemod:sounds/dashforward-1"), player, false, 1.1f, 0.2f);
                        }
                        else if (rndInt == 3)
                        {
                            api.World.PlaySoundFor(new AssetLocation("vsmovemod:sounds/dashforward-2"), player, false, 1.1f, 0.2f);
                        }

                    }
                    else if (stamina > drainStaminaDash && dashCountdown < 1 && !player.Entity.Controls.Backward && player.Entity.Controls.Right)
                    {
                        player.Entity.StartAnimation("dashright");
                        stamina = stamina -= drainStaminaDash;
                        dashCountdown = maxDashCountdown;
                        staminaRegenDelay = maxStaminaRegenDelay;
                        dashTime = 0;
                        motionStopedAfterDash = false;

                        api.World.PlaySoundFor(new AssetLocation("vsmovemod:sounds/dash"), player, false, 2, 0.3f);
                    }
                    else if (stamina > drainStaminaDash && dashCountdown < 1 && !player.Entity.Controls.Backward && player.Entity.Controls.Left)
                    {
                        player.Entity.StartAnimation("dashleft");
                        stamina = stamina -= drainStaminaDash;
                        dashCountdown = maxDashCountdown;
                        staminaRegenDelay = maxStaminaRegenDelay;
                        dashTime = 0;
                        motionStopedAfterDash = false;

                        api.World.PlaySoundFor(new AssetLocation("vsmovemod:sounds/dash"), player, false, 2, 0.3f);
                    }
                    else if (stamina > drainStaminaDash && dashCountdown < 1 && player.Entity.Controls.Backward)
                    {
                        player.Entity.StartAnimation("backflip");
                        stamina = stamina -= drainStaminaDash;
                        dashCountdown = maxDashCountdown;
                        staminaRegenDelay = maxStaminaRegenDelay;
                        dashTime = 0;
                        motionStopedAfterDash = false;

                        api.World.PlaySoundFor(new AssetLocation("vsmovemod:sounds/backflip"), player, false, 4, 0.2f);

                    }
                    else if (stamina < drainStaminaDash)
                    {
                        stamina = 0;
                        staminaRegenDelay = maxStaminaRegenDelay;
                    }
                }

                // Slide start
                if (player.Entity.Controls.Sprint && player.Entity.Controls.TriesToMove && (api.Input.MouseButton.Middle == true) && dashCountdown < 1 && !isCrawling)
                {
                    if (stamina > drainStaminaDash && dashCountdown < 1)
                    {
                        player.Entity.Controls.StopAllMovement();
                        if (!inGliderMode)
                        {
                            isCrawling = true;
                        }
                        stamina = stamina -= drainStaminaDash;
                        dashCountdown = maxDashCountdown;
                        staminaRegenDelay = maxStaminaRegenDelay;
                        slideTime = 0;
                    }
                }

                // Slide-Movement /  50ms
                if (slideTime < 10)
                {
                    if (!player.Entity.CollidedVertically)
                    {
                        player.Entity.Pos.Motion = player.Entity.Pos.Motion * 1;
                        slideTime++;
                    }
                    else
                    {
                        player.Entity.Pos.Motion = player.Entity.Pos.Motion * 3;
                        slideTime++;
                    }
                }

                // Dash-Movement /  50ms
                if (dashTime < 5)
                {
                    if (!player.Entity.CollidedVertically && !player.Entity.Controls.Backward)
                    {
                        player.Entity.Pos.Motion = player.Entity.Pos.Motion * 1.10;
                        dashTime++;
                    }
                    else if (!player.Entity.CollidedVertically && player.Entity.Controls.Backward)
                    {
                        player.Entity.Pos.Motion = player.Entity.Pos.Motion * 1;
                        player.Entity.Pos.Motion.Y = player.Entity.Pos.Motion.Y * 0.6;
                        dashTime++;
                    }
                    else
                    {
                        player.Entity.Pos.Motion = player.Entity.Pos.Motion * 2;
                        dashTime++;
                    }
                }

                // Wall Jump
                if (player.Entity.Controls.IsClimbing && player.Entity.Controls.RightMouseDown && !player.Entity.Controls.Backward && player.Entity.Controls.Left)
                {
                    if (stamina > drainStaminaDash && dashCountdown < 1)
                    {
                        player.Entity.Properties.CanClimbAnywhere = false;
                        player.Entity.Pos.Motion = player.Entity.Pos.Motion * 6;
                        stamina = stamina -= drainStaminaDash;
                        staminaRegenDelay = maxStaminaRegenDelay;
                        dashCountdown = maxDashCountdown;

                        player.Entity.StartAnimation("dashleft");
                    }
                }
                else if (player.Entity.Controls.IsClimbing && player.Entity.Controls.RightMouseDown && !player.Entity.Controls.Backward && player.Entity.Controls.Right)
                {
                    if (stamina > drainStaminaDash && dashCountdown < 1)
                    {
                        player.Entity.Properties.CanClimbAnywhere = false;
                        player.Entity.Pos.Motion = player.Entity.Pos.Motion * 6;
                        stamina = stamina -= drainStaminaDash;
                        staminaRegenDelay = maxStaminaRegenDelay;
                        dashCountdown = maxDashCountdown;

                        player.Entity.StartAnimation("dashright");
                    }
                }
                else if (player.Entity.Controls.IsClimbing && player.Entity.Controls.RightMouseDown && player.Entity.Controls.Backward)
                {
                    if (stamina > drainStaminaDash && dashCountdown < 1)
                    {
                        player.Entity.Properties.CanClimbAnywhere = false;
                        player.Entity.Pos.Motion = player.Entity.Pos.Motion * 1.5;
                        stamina = stamina -= drainStaminaDash;
                        staminaRegenDelay = maxStaminaRegenDelay;
                        dashCountdown = maxDashCountdown;
                    }
                }

                // Charging jumpforce
                if (player.Entity.Controls.Sneak && player.Entity.CollidedVertically && !player.Entity.Controls.TriesToMove && !isCrawling && !(player.Entity.CurrentControls == (EnumEntityActivity.FloorSitting | EnumEntityActivity.Idle)))
                {
                    if (((maxChargeJump - chargeJump) / 10 * drainStaminaChargeJump) < stamina + 1 && chargeJump <= maxChargeJump)
                    {
                        chargeJump += 10;
                        stamina -= drainStaminaChargeJump;
                    }
                }
                else if (chargeJump > 0)
                {
                    chargeJump -= 5;
                    if (chargeJump < 0)
                    {
                        chargeJump = 0;
                    }
                }

                // Charge Jump
                if (chargeJump >= 100)
                {
                    Vintagestory.API.Config.GlobalConstants.BaseJumpForce = 12f;
                    staminaRegenDelay = maxStaminaRegenDelay;
                    player.Entity.StartAnimation("powerjump");
                }
                else
                {
                    Vintagestory.API.Config.GlobalConstants.BaseJumpForce = 8.2f;
                    player.Entity.StopAnimation("powerjump");
                }

                // Check if block at playerPos is climable (ladder)
                if (checkBlockatPlayerPos.Climbable)
                {
                    isClimbable = true;
                }
                else
                {
                    isClimbable = false;
                }

                // Stamina drain while climbing
                if (player.Entity.Controls.IsClimbing && stamina > 0 && (player.Entity.Controls.TriesToMove || player.Entity.Controls.Sneak || player.Entity.Controls.Jump))
                {
                    // Sound random climbing
                    if (soundClimbinglCountdown < 10)
                    {
                        Random rnd = new Random();
                        api.World.PlaySoundFor(new AssetLocation("vsmovemod:sounds/climb"), player, false, 5, 0.2f);
                        //Playback-rate randomizer
                        soundClimbinglCountdown = rnd.Next(30, 60);
                    }

                    float staminaNewValue = (stamina -= drainStaminaClimbing);
                    if (isClimbable == false)
                    {
                        if (staminaNewValue >= 0)
                        {
                            stamina = staminaNewValue;
                        }
                        else
                        {
                            stamina = 0;
                        }
                    }
                    else if (isClimbable == true)
                    {
                        stamina += drainStaminaClimbing;
                    }
                }
                else if (player.Entity.Controls.IsClimbing && stamina > 0 && !(player.Entity.Controls.TriesToMove || player.Entity.Controls.Sneak || player.Entity.Controls.Jump))
                {
                    float staminaNewValue = (stamina -= (drainStaminaClimbing * drainStaminaIdleModificator));
                    if (isClimbable == false)
                    {
                        if (staminaNewValue >= 0)
                        {
                            stamina = staminaNewValue;
                        }
                        else
                        {
                            stamina = 0;
                        }
                    }
                    else if (isClimbable == true)
                    {
                        stamina += (drainStaminaClimbing * drainStaminaIdleModificator);
                    }
                }

                // Ability to climb
                if (!player.Entity.Controls.IsClimbing)
                {
                    if (player.Entity.Controls.Sneak && stamina > drainStaminaClimbing && player.InventoryManager.ActiveHotbarSlot.Empty)
                    {
                        player.Entity.Properties.CanClimbAnywhere = true;
                    }
                    else
                    {
                        player.Entity.Properties.CanClimbAnywhere = false;
                        letGoClimbing = false;
                    }
                }
                else
                {
                    if (stamina > 0 && player.InventoryManager.ActiveHotbarSlot.Empty && !letGoClimbing)
                    {
                        player.Entity.Properties.CanClimbAnywhere = true;
                    }
                    else if (isClimbable == true)
                    {
                        player.Entity.Properties.CanClimbAnywhere = true;
                    }
                    else
                    {
                        Block BlockatPlayerZMinus = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y, (int)player.Entity.Pos.Z - 1);
                        Block BlockatPlayerZPlus = api.World.BlockAccessor.GetBlock((int)player.Entity.Pos.X, (int)player.Entity.Pos.Y, (int)player.Entity.Pos.Z + 1);

                        // Sound loose grip (no stamina climbing)
                        if (!isClimbable && !aboveBlock.Climbable && !BlockatPlayerZMinus.Climbable && !BlockatPlayerZPlus.Climbable)
                        {
                            api.World.PlaySoundFor(new AssetLocation("vsmovemod:sounds/loosegrip"), player, false, 35, 0.2f);
                        }
                        player.Entity.Properties.CanClimbAnywhere = false;
                        letGoClimbing = false; 
                    }
                }

                // Fix no ladderup animation when holding sprint
                if (player.Entity.Controls.IsClimbing && (player.Entity.Controls.TriesToMove || player.Entity.Controls.Jump) && !player.Entity.Controls.Sneak)
                {
                    player.Entity.StartAnimation("climbup");
                }
                else if (player.Entity.Controls.IsClimbing && !(player.Entity.Controls.TriesToMove || player.Entity.Controls.Jump) && !player.Entity.Controls.Sneak)
                {
                    player.Entity.StopAnimation("climbup");
                }
                else if (!player.Entity.Controls.IsClimbing && !player.Entity.Controls.Sneak)
                {
                    player.Entity.StopAnimation("climbup");
                }

                // Stop crouch animation & allow movement while climbing (Crouch-climb fix)
                if (player.Entity.Controls.IsClimbing)
                {
                    player.Entity.StopAnimation("gleiten");
                    player.Entity.StopAnimation("liegen");
                    api.Input.InWorldAction += Input_InWorldAction1;
                }

                // Stepheight while sprinting
                if (player.Entity.Controls.Sprint && !noStep)
                {
                    if (walkStep || sprintStep)
                    {
                        player.Entity.GetBehavior<EntityBehaviorControlledPhysics>().stepHeight = 1.0F;
                    }
                }
                else
                {
                    if (!walkStep)
                    {
                        player.Entity.GetBehavior<EntityBehaviorControlledPhysics>().stepHeight = 0.9F;
                    }
                    else
                    {
                        player.Entity.GetBehavior<EntityBehaviorControlledPhysics>().stepHeight = 1.0F;
                    }
                }

                //Delay Stamina Regen when Stamina < 1 (0)
                if (stamina < 1 && staminaRegenIsDelayed == false)
                {
                    staminaRegenDelay = maxStaminaRegenDelay;
                    staminaRegenIsDelayed = true;
                    generalAudioDelay = 15;
                }

                // Sound no Stamina
                if (stamina < 1 && generalAudioDelay < 10)
                {
                    api.World.PlaySoundFor(new AssetLocation("vsmovemod:sounds/nostam"), player, false, 0, 0.2f);
                    generalAudioDelay = 80;
                }             

                // Stamina Regen
                if (stamina < maxStamina && !player.Entity.Controls.IsClimbing && player.Entity.CollidedVertically && !player.Entity.Controls.Sneak && staminaRegenDelay < 1)
                {
                    float staminaNewValue;
                    staminaNewValue = (stamina += staminaRegenValue);

                    if (staminaNewValue > maxStamina)
                    {
                        stamina = maxStamina;
                    }
                    else
                    {
                        stamina += staminaRegenValue;
                    }

                    // Sound recover Stamina
                    if (generalAudioDelay < 60 && staminaRegenIsDelayed)
                    {
                        Random rnd = new Random();
                        api.World.PlaySoundFor(new AssetLocation("vsmovemod:sounds/recover"), player, false, 0, 0.2f);
                        ////Playback-rate randomizer
                        generalAudioDelay = 80;
                    }

                    if (stamina == maxStamina)
                    {
                        staminaRegenIsDelayed = false;
                    }
                }

                // ========= Experimental =========

                //// Gear Change Event
                //player.Entity.GearInventory.SlotModified += GearInventory_SlotModified;
                //if (slotChanged)
                //{
                //    slotChanged = false;
                //    player.ShowChatNotification("Item Slot changed");
                //}

                //// Check for specific item in active slot
                //Item checkItem = api.World.GetItem(new AssetLocation("bow-long"));
                //if (!player.InventoryManager.ActiveHotbarSlot.Empty)
                //{
                //    if (player.InventoryManager.ActiveHotbarSlot.Itemstack.Item == checkItem)
                //    {
                //        player.ShowChatNotification("Item in hand is a bow-long");
                //    }
                //}
                //else           // null Exception catch
                //{
                //    player.ShowChatNotification("No item in hand");
                //}

                ////Check for specific wearable
                //var gearInv = player.Entity.GearInventory;
                //Item checkItemWearable = api.World.GetItem(new AssetLocation("game:clothes-head-silver-diadem"));

                //foreach (var slot in gearInv)
                //{
                //    ItemWearable item;
                //    if (slot.Empty || (item = slot.Itemstack.Collectible as ItemWearable) == null) continue; // null Exception catch
                //    if ((item = slot.Itemstack.Collectible as ItemWearable) == checkItemWearable)
                //    {
                //        player.ShowChatNotification("You are wearing the mystic item");
                //    }
                //}

                //// Take damage
                //DamageSource fallDamage = (new DamageSource()
                //{
                //    Source = EnumDamageSource.Fall,
                //    Type = EnumDamageType.Gravity
                //});

                //if (isDoingSomething)
                //{
                //    player.Entity.ShouldReceiveDamage(fallDamage, 1);       
                //}

                //// Falldamage
                //if (((player.Entity.CurrentControls == (EnumEntityActivity.Fall | EnumEntityActivity.Idle)) || (player.Entity.CurrentControls == (EnumEntityActivity.Fall | EnumEntityActivity.Move))))
                //{
                //    isFalling = true;

                //    if (isFalling && player.Entity.Pos.Motion.Y < -0.19)
                //    {
                //        isFallingFromLow = true;

                //        if (isFallingFromLow && player.Entity.CollidedVertically)
                //        {

                //            // Define falldamage
                //            DamageSource fallDamage = (new DamageSource()
                //            {
                //                Source = EnumDamageSource.Unknown,
                //                Type = EnumDamageType.Gravity
                //            }); ;

                //            int damageAmount = 100000000;
                //            player.Entity.ShouldReceiveDamage(fallDamage, damageAmount);

                //            isFalling = false;
                //            isFallingFromLow = false;
                //        }
                //    }
                //}
                //else
                //{
                //    isFalling = false;
                //}

            }, 50); //50 ms/tick


        }

        // Hotkey stand-up Method
        private bool StandUp(KeyCombination comb)
        {
            if (isCrawling)
            {
                MoveMod.toggleStandup = true;
                return true;
            }

            return true;
        }

        // Hotkey start-crawl Method
        private bool StartCrawl(KeyCombination comb)
        {
            if (!isCrawling)
            {
                MoveMod.toggleCrawl = true;
                return true;
            }

            return true;
        }

        // Hotkey toggle-glider Method
        private bool UseGlider(KeyCombination comb)
        {
            if (!playerIsClimbing)
            {
                if (!toggleGlider)
                {
                    toggleGlider = true;
                    return true;
                }
                else if (toggleGlider)
                {
                    toggleGlider = false;
                    return true;
                }
            }
            else if (playerIsClimbing)
            {
                letGoClimbing = true;
                return true;
            }
            return true;
        }


        // Eventhandler on Inventory-Gear changed
        //private void GearInventory_SlotModified(int t1)
        //{
        //    slotChanged = true;
        //}


        // EventHandler PlayerAction - Prevent Sneaking / climbing down
        private void Input_InWorldAction3(EnumEntityAction action, bool on, ref EnumHandling handled)
        {
            //Prevent sneak
            if (action == EnumEntityAction.Sneak)
            {
                handled = EnumHandling.PreventDefault;
            }
        }

        // EventHandler PlayerAction - Prevent sneaking / Sprinting
        private void Input_InWorldAction4(EnumEntityAction action, bool on, ref EnumHandling handled)
        {
            //Prevent sneak
            if (action == EnumEntityAction.Sneak)
            {
                handled = EnumHandling.PreventDefault;
            }
            //Prevent sprint
            if (action == EnumEntityAction.Sprint)
            {
                handled = EnumHandling.PreventDefault;
            }
        }

        // EventHandler PlayerAction - Allow all movement & Prevent Sprint,Jump
        private void Input_InWorldAction2(EnumEntityAction action, bool on, ref EnumHandling handled)
        {
            //Prevent sprint / jump
            if (action == EnumEntityAction.Sprint)
            {
                handled = EnumHandling.PreventDefault;
            }
            else if (action == EnumEntityAction.Jump)
            {
                handled = EnumHandling.PreventDefault;
            }

            //allow all movement directions
            else if (action == EnumEntityAction.Forward)
            {
                handled = EnumHandling.PassThrough;
            }
            else if (action == EnumEntityAction.Left)
            {
                handled = EnumHandling.PassThrough;
            }
            if (action == EnumEntityAction.Right)
            {
                handled = EnumHandling.PassThrough;
            }
            else if (action == EnumEntityAction.Backward)
            {
                handled = EnumHandling.PassThrough;
            }
        }

        // EventHandler PlayerAction - Prevent all directions & Sprint,Jump
        private void Input_InWorldAction(EnumEntityAction action, bool on, ref EnumHandling handled)
        {
            //Prevent sprint / jump
            if (action == EnumEntityAction.Sprint)
            {
                handled = EnumHandling.PreventDefault;
            }
            else if (action == EnumEntityAction.Jump)
            {
                handled = EnumHandling.PreventDefault;
            }

            //prevent all movement directions
            if (action == EnumEntityAction.Forward)
            {
                handled = EnumHandling.PreventDefault;
            }
            else if (action == EnumEntityAction.Left)
            {
                handled = EnumHandling.PreventDefault;
            }
            if (action == EnumEntityAction.Right)
            {
                handled = EnumHandling.PreventDefault;
            }
            else if (action == EnumEntityAction.Backward)
            {
                handled = EnumHandling.PreventDefault;
            }

        }

        // EventHandler PlayerAction - Allow all movement
        private void Input_InWorldAction1(EnumEntityAction action, bool on, ref EnumHandling handled)
        {
            if (action == EnumEntityAction.Sprint)
            {
                handled = EnumHandling.PassThrough;
            }
            else if (action == EnumEntityAction.Jump)
            {
                handled = EnumHandling.PassThrough;
            }
            else if (action == EnumEntityAction.Sneak)
            {
                handled = EnumHandling.PassThrough;
            }

            //allow all movement directions
            else if (action == EnumEntityAction.Forward)
            {
                handled = EnumHandling.PassThrough;
            }
            else if (action == EnumEntityAction.Left)
            {
                handled = EnumHandling.PassThrough;
            }
            else if (action == EnumEntityAction.Right)
            {
                handled = EnumHandling.PassThrough;
            }
            else if (action == EnumEntityAction.Backward)
            {
                handled = EnumHandling.PassThrough;
            }
        }
    }

    // Stamina GUI
    // -----------------------------------------------------------------------------------
    public class HudOverlaySample : ModSystem
    {
        ICoreClientAPI capi;
        StaminaProgressBarRenderer renderer;

        public override bool ShouldLoad(EnumAppSide side)
        {
            return side == EnumAppSide.Client;
        }

        public override void StartClientSide(ICoreClientAPI api)
        {
            this.capi = api;
            renderer = new StaminaProgressBarRenderer(api);

            api.Event.RegisterRenderer(renderer, EnumRenderStage.Ortho);
        }
    }

    public class StaminaProgressBarRenderer : IRenderer
    {
        MeshRef whiteRectangleRef;
        MeshRef progressQuadRef;
        ICoreClientAPI capi;

        Matrixf mvMatrix = new Matrixf();

        public double RenderOrder { get { return 0; } }

        public int RenderRange { get { return 10; } }

        public StaminaProgressBarRenderer(ICoreClientAPI api)
        {
            this.capi = api;

            // This will get a line loop with vertices inside [-1,-1] till [1,1]
            MeshData rectangle = LineMeshUtil.GetRectangle(ColorUtil.WhiteArgb);
            whiteRectangleRef = api.Render.UploadMesh(rectangle);

            // This will get a quad with vertices inside [-1,-1] till [1,1]
            progressQuadRef = api.Render.UploadMesh(QuadMeshUtil.GetQuad());
        }

        public void OnRenderFrame(float deltaTime, EnumRenderStage stage)
        {
            IShaderProgram curShader = capi.Render.CurrentActiveShader;

            Vec4f color = new Vec4f(1, 1, 1, 1);

            if (MoveMod.playerIsClimbing == true || MoveMod.stamina < MoveMod.maxStamina)
            {
                // Render rectangle
                curShader.Uniform("rgbaIn", color);
                curShader.Uniform("extraGlow", 0);
                curShader.Uniform("applyColor", 0);
                curShader.Uniform("tex2d", 0);
                curShader.Uniform("noTexture", 1f);

                float width = 100;

                mvMatrix
                    .Set(capi.Render.CurrentModelviewMatrix)
                    .Translate(capi.Render.FrameWidth / 2 - 50, capi.Render.FrameHeight / 8, 0)
                    .Scale(width, 20, 0)
                    .Translate(0.5f, 0.5f, 0)
                    .Scale(0.5f, 0.5f, 0)
                ;
                curShader.UniformMatrix("projectionMatrix", capi.Render.CurrentProjectionMatrix);
                curShader.UniformMatrix("modelViewMatrix", mvMatrix.Values);

                capi.Render.RenderMesh(whiteRectangleRef);
            }


            // Render progress bar
            if (MoveMod.playerIsClimbing == true || MoveMod.stamina < MoveMod.maxStamina)
            {
                if (MoveMod.stamina >= MoveMod.maxStamina)
                {
                    float width = 100;
                    mvMatrix
                        .Set(capi.Render.CurrentModelviewMatrix)
                        .Translate(capi.Render.FrameWidth / 2 - 50, capi.Render.FrameHeight / 8, 0)
                        .Scale(width, 20, 0)
                        .Translate(0.5f, 0.5f, 0)
                        .Scale(0.5f, 0.5f, 0)
                    ;
                }
                else
                {
                    float width = ((MoveMod.stamina) / MoveMod.maxStamina) * 100;
                    mvMatrix
                        .Set(capi.Render.CurrentModelviewMatrix)
                        .Translate(capi.Render.FrameWidth / 2 - 50, capi.Render.FrameHeight / 8, 0)
                        .Scale(width, 20, 0)
                        .Translate(0.5f, 0.5f, 0)
                        .Scale(0.5f, 0.5f, 0)
                    ;
                }
            }

            if (MoveMod.playerIsClimbing == true || MoveMod.stamina < MoveMod.maxStamina)
            {
                curShader.UniformMatrix("projectionMatrix", capi.Render.CurrentProjectionMatrix);
                curShader.UniformMatrix("modelViewMatrix", mvMatrix.Values);

                capi.Render.RenderMesh(progressQuadRef);
            }

        }

        public void Dispose()
        {
            capi.Render.DeleteMesh(whiteRectangleRef);
            capi.Render.DeleteMesh(progressQuadRef);
        }
    }
}
